open GdkEvent

module C = Cairo
module CaMa = C.Matrix

let ( *| ) = CaMa.multiply
let pi_2 = 8.0 *. atan 1.0

type state =
  {
    adding : (float * float * float) option;
    circles : (float * float * float) array;
    dragging : int option;
    pan : (bool * float * float) option;
    matrix : C.matrix;
  }

let init_state =
  {
    adding = None;
    circles = [| |];
    dragging = None;
    pan = None;
    matrix = CaMa.init_identity
  }

let init_width = 1024
let init_height = 768

let win_title = "Circle builder demo"

let cmd_line_handler s _ = s

let draw_circle cr x y rad fill =
  C.arc cr ~xc:x ~yc:y ~radius:rad ~angle1:0.0 ~angle2:pi_2;
  if fill then C.fill cr else C.stroke cr

let keyboard_callback state key =
  match get_type key with
    | `KEY_PRESS when Key.string key = "q"  ->
        Printf.printf "Quit!\n";
        GMain.quit ();
        state, true
    | _ -> state, false

exception Break

let dist x0 y0 x1 y1 =
  sqrt ((x0 -. x1) ** 2. +. (y0 -. y1) ** 2.)

let is_inside_circle circles x y =
  let inside = ref None in
  let len = Array.length circles in
    try
      for i = 0 to len-1
      do
        if !inside <> None then raise Break
        else
          let xc, yc, radius = circles.(i) in
            if dist xc yc x y < radius then inside := Some i
      done;
        !inside
    with
      | Break -> !inside

let get_usr_point state x y =
  let phy_p = {C.x = x; C.y = y} in
  let usr_p = CaMa.transform_point (CaMa.invert state.matrix) phy_p in
    usr_p.C.x, usr_p.C.y

let button_1 state b_ev =
  let x, y = get_usr_point state (Button.x b_ev) (Button.y b_ev) in
    match get_type b_ev with
      | `BUTTON_PRESS ->
          let inside = is_inside_circle state.circles x y in (
            match inside with
              | None -> {state with adding = Some (x, y, 0.0)}, true
              | Some i -> {state with dragging = inside}, false
          )
      | `BUTTON_RELEASE -> (
          match state.adding with
            | None -> {state with dragging = None}, true
            | Some (xc, yc, _) ->
                let radius = dist x y xc yc in
                  {state with
                    circles = Array.append state.circles [|xc, yc, radius|];
                    adding = None;
                    dragging = None;
                  }, true
        )
      | _ -> state, false

let button_2 state b_ev =
  let x, y = Button.x b_ev, Button.y b_ev in
    match get_type b_ev with
      | `BUTTON_PRESS ->
          {state with pan = Some (false, x, y)}, true
      | `BUTTON_RELEASE -> (
          match state.pan with
            | None -> state, false
            | Some (moving, x_press, y_press) ->
                if moving = false
                then {state with matrix = CaMa.init_identity; pan = None}, true
                else {state with pan = None}, true
        )
      | _ -> state, false

let button_3 state b_ev =
  let x, y = get_usr_point state (Button.x b_ev) (Button.y b_ev) in
    match get_type b_ev with
      | `BUTTON_PRESS ->
          let inside = is_inside_circle state.circles x y in (
            match inside with
              | None -> state, false
              | Some i ->
                  let n = (Array.length state.circles) in
                  let new_circles = Array.make (n - 1) (0., 0., 0.) in
                  let ind = ref 0 in
                    for cont = 0 to n-1
                    do
                      if cont <> i
                      then begin
                        new_circles.(!ind) <- state.circles.(cont);
                        incr ind
                      end
                    done;
                    {state with circles = new_circles}, true
          )
      | _ -> state, false

let pointer_buttons_callback state b_ev =
    match Button.button b_ev with
      | 1 -> button_1 state b_ev
      | 2 -> button_2 state b_ev
      | 3 -> button_3 state b_ev
      | _ -> state, false

let pointer_motion_callback state m_ev =
  let x, y = get_usr_point state (Motion.x m_ev) (Motion.y m_ev) in
    match state.adding, state.dragging, state.pan with
      | Some (xc, yc, _), _, _ ->
        let radius = dist x y xc yc in
          {state with adding = Some (xc, yc, radius)}, true
      | _, None, None -> state, false
      | _, None, Some (_, start_x, start_y) ->
        let x, y = Motion.x m_ev, Motion.y m_ev in
        let tr = CaMa.init_translate (x -. start_x) (y -. start_y) in
          {state with matrix = state.matrix *| tr; pan = Some (true, x, y)}, true
      | _, Some i, _ ->
          let _, _, rad = state.circles.(i) in
          let x, y = get_usr_point state (Motion.x m_ev) (Motion.y m_ev) in
          state.circles.(i) <- (x, y, rad); state, true

let pointer_scroll_callback state sc_ev =
  let scal_fact =
    match Scroll.direction sc_ev with | `UP -> 1.1 | `DOWN -> 0.9 | _ -> 0.0 in
  let x_abs, y_abs = Scroll.x sc_ev, Scroll.y sc_ev in
  let m1 = CaMa.init_translate (-.x_abs) (-.y_abs) in
  let m2 = CaMa.init_scale scal_fact scal_fact in
  let m3 = CaMa.init_translate x_abs y_abs in
  let m = state.matrix *| m1 *| m2 *| m3 in
    {state with matrix = m}, true

let repaint state cr width height =
  C.set_antialias cr C.ANTIALIAS_SUBPIXEL;
  C.transform cr state.matrix;
  match state.dragging, state.adding with
    | Some i, _ ->
        Array.iteri (fun ind (xc, yc, rad) ->
          let fill = if i <> ind then false else true in
            draw_circle cr xc yc rad fill) state.circles;
          state, true
    | _, Some (xc, yc, rad) ->
      Array.iter (fun (x, y, rad) -> draw_circle cr x y rad false) state.circles;
      draw_circle cr xc yc rad true; state, true
    | _, None ->
      Array.iter (fun (x, y, rad) -> draw_circle cr x y rad false) state.circles;
      state, true

