module type INTERACTOR =
sig
  type state

  val init_state : state

  val win_title : string
  val init_width : int
  val init_height : int

  val cmd_line_handler : state -> string array -> state
  val keyboard_callback : state -> GdkEvent.Key.t -> state * bool
  val pointer_buttons_callback : state -> GdkEvent.Button.t -> state * bool
  val pointer_motion_callback : state -> GdkEvent.Motion.t -> state * bool
  val pointer_scroll_callback : state -> GdkEvent.Scroll.t -> state * bool

  val repaint : state -> Cairo.t -> int -> int -> state * bool
end

module type TOY =
sig
  val run : unit -> unit
end

module Make (I : INTERACTOR) =
struct
  let w = GWindow.window  ~title:I.win_title
    ~width:I.init_width
    ~height:I.init_height ()
  let b = GPack.vbox ~spacing:0 ~border_width:0  ~packing:w#add ()
  let f = GBin.frame ~shadow_type:`IN ~packing:(b#pack ~expand:true ~fill:true) ()
  let area = GMisc.drawing_area ~width:200 ~height:200 ~packing:f#add ()

  let current_state = ref I.init_state

  let redraw w expose_ev =
    let cr = Cairo_lablgtk.create w#misc#window in
    let { Gtk.width = width; Gtk.height = height } = w#misc#allocation in
      Cairo.save cr;
      let new_state, res = I.repaint !current_state cr width height in
        current_state := new_state;
        Cairo.restore cr;
        res

  let wrap_event callback ev =
    let new_state, res = callback !current_state ev in
      current_state := new_state;
      if res = true
      then GtkBase.Widget.queue_draw w#as_widget;
        res

  let run () =
    current_state := I.cmd_line_handler !current_state Sys.argv;

    w#event#add [ `BUTTON_PRESS; `BUTTON_RELEASE; `POINTER_MOTION ];
    ignore (w#connect#destroy GMain.quit);
    ignore (w#event#connect#key_press ~callback:(wrap_event I.keyboard_callback));
    ignore (w#event#connect#key_release ~callback:(wrap_event I.keyboard_callback));
    ignore (w#event#connect#button_press ~callback:(wrap_event I.pointer_buttons_callback));
    ignore (w#event#connect#button_release ~callback:(wrap_event I.pointer_buttons_callback));
    ignore (w#event#connect#motion_notify ~callback:(wrap_event I.pointer_motion_callback));
    ignore (w#event#connect#scroll ~callback:(wrap_event I.pointer_scroll_callback));

    ignore (area#event#connect#expose (redraw area));
    w#show ();
    GMain.main ()
end

